use hyper::{Body, Request as HyperRequest};
use typemap::SendMap;
use tokio_core::reactor::Remote;

use RawResponse;
use errors::HandleError;
use outcome::Outcome;
use request::{self, Request};
use response::{Response, Responder};
use router::regex::OwnedCaptures;


/// A context during handling.
///
/// It contains an HTTP request, HTTP response to return,
/// and a typemap in order to share variables between middlewares.
pub struct Context {
    /// Incoming request contains URI, method, headers, and body
    pub req: Request,
    /// Captured parameters from router
    pub params: OwnedCaptures,
    /// Request body
    pub body: Option<Body>,
    /// Handle of event loop
    pub handle: Remote,
    /// Extension map to share values between handlers/middlewares
    pub ext: SendMap,
    pub(crate) on_send: Vec<Box<Fn(&mut RawResponse) + Send + 'static>>,
}

impl Context {
    pub(crate) fn new(req: HyperRequest, params: OwnedCaptures, handle: Remote) -> Self {
        let (req, body) = request::reconstruct(req);
        Context {
            req,
            params,
            body: Some(body),
            handle,
            ext: SendMap::custom(),
            on_send: Vec::new(),
        }
    }

    pub fn respond<R: Responder>(self, responder: R) -> Outcome<Response> {
        let mut res = match responder.respond(&self.req) {
            Ok(res) => res,
            Err(err) => return Outcome::halt(err),
        };
        for callback in self.on_send {
            callback(&mut res);
        }
        Outcome::continue_(Response { res })
    }

    pub fn next(self) -> Outcome<Context> {
        Outcome::continue_(self)
    }

    pub fn halt<T>(self, err: HandleError) -> Outcome<T> {
        Outcome::halt(err)
    }

    /// Register a callback function called after construct response.
    pub fn on_send<F>(&mut self, on_send: F) -> &mut Self
    where
        F: Fn(&mut RawResponse) + Send + 'static,
    {
        self.on_send.push(Box::new(on_send));
        self
    }
}
